package common;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.StringReader;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.util.ArrayList;

public class FileStream implements JobStreamI {
	private static final int BUFFER_SIZE=64*1024 /*Blu-Ray ECC Size*/ * 256;
	private Logger logger=Logger.getLogger(this.getClass());
	
	private Job job;
	
	protected File files[];
	private int currentIndex=-1;
	private int chunkSize;
	private int readSize=BUFFER_SIZE;
	private long filePosition;
	private long streamPosition;
	private long streamSize;
	private MappedByteBuffer fileBuffer;
	private MappedByteBuffer readAheadBuffer;
	private long readAheadPosition=-1;
	
	private FileChannel fileChannel;
	
	/**
	 * Expects a list of /n separated file names
	 * @param fileName
	 */
	public static File[] getFiles(String fileNames) {
		ArrayList<File> fileList=new ArrayList<File>(20);
		String fileName;
		try {
			for (BufferedReader reader=new BufferedReader(new StringReader(fileNames));(fileName=reader.readLine()) != null;) {
				if (fileName.trim().length() > 0) {
					File file=new File(fileName);
					if (file.exists()) {
						fileList.add(file);
					} else {
						throw new FileNotFoundException(fileName);
					}
				}
			}
		} catch (IOException e) {
			// Can't happen
			e.printStackTrace();
		}
		File files[]=new File[fileList.size()];
		fileList.toArray(files);
		return files;
	}
	
	public FileStream(Job job, File files[]) {
		this.job=job;
		this.files=files;
		reset();
	}
	public void reset() {
		fileChannel=null;
		fileBuffer=null;
		filePosition=0;
		readAheadBuffer=null;
		readAheadPosition=-1;
		streamPosition=0;
		currentIndex=-1;
		streamSize=0;
		if (job!=null) {
			setJob(job);
		}
	}
	
	private int ensureBuffer(int len) throws IOException {
		int remaining=fileBuffer.capacity()-fileBuffer.position();
		if (remaining < len) {
			filePosition -= remaining;
			readFile();
			//This deals with oddball file lengths
			len=Math.min(len, fileBuffer.capacity()-fileBuffer.position());
		}
		return len;
	}
	
	/*public int readFully(final byte buffer[], final int offset, int len) throws IOException {
		len=ensureBuffer(len);
		fileBuffer.get(buffer, offset, len);
		return len;
	}*/
	
	public ByteBuffer slice(int len) throws IOException {
		len=ensureBuffer(len);
		
		int oldLimit=fileBuffer.limit();
		fileBuffer.limit(fileBuffer.position()+len);
		ByteBuffer slice=fileBuffer.slice();
		fileBuffer.limit(oldLimit);
		fileBuffer.position(fileBuffer.position()+len);
		return slice;
	}
	public final long getAbsoluteFilePosition() {
		return filePosition - fileBuffer.capacity() + fileBuffer.position();
	}
	public void open() throws IOException {
		if (currentIndex >= 0) {
			streamPosition += fileChannel.size();
			closeFile();
		}
		if (currentIndex++ < files.length) {
			logger.info("Processing file: " + files[currentIndex]);
			RandomAccessFile fileIn=new RandomAccessFile(files[currentIndex],"r"); 
			fileChannel=fileIn.getChannel();
			//System.out.println("Opened: " + files[currentIndex]);
			filePosition=0;
			readFile();
			job.setDescription(files[currentIndex].getAbsolutePath());
		}
	}
	
	private void readFile() throws IOException {
		if (filePosition < fileChannel.size()) {
			long bytesToRead;
			if (filePosition==readAheadPosition) {
				bytesToRead=readAheadBuffer.limit();
				fileBuffer=readAheadBuffer;
			} else {
				bytesToRead=Math.min(fileChannel.size()-filePosition, readSize);
				fileBuffer=DirectMemoryManager.mapMemory(fileChannel, FileChannel.MapMode.READ_ONLY, filePosition, bytesToRead);
			}
			job.notifyListeners();
			filePosition+=bytesToRead;
			if (filePosition < fileChannel.size()) {
				//Read ahead
				bytesToRead=Math.min(fileChannel.size()-filePosition, readSize);
				readAheadBuffer=DirectMemoryManager.mapMemory(fileChannel, FileChannel.MapMode.READ_ONLY, filePosition, bytesToRead);
				readAheadPosition=filePosition;
			} else {
				readAheadPosition=-1;
				readAheadBuffer=null;
			}
			job.setPosition((int)((streamPosition + getAbsoluteFilePosition())/Job.KB));
		} else {
			open();
		}
	}

	public byte peek(int offset) throws IOException {
		int remaining=fileBuffer.capacity()-fileBuffer.position();
		if (remaining <= offset) {
			filePosition -= remaining;
			readFile();
		}
		return fileBuffer.get(fileBuffer.position() + offset);
	}
	public int skip(int bytes) throws IOException {
		if (bytes <= fileBuffer.remaining()) {
			fileBuffer.position(fileBuffer.position()+bytes);
		} else {
			filePosition+=bytes-fileBuffer.remaining();
			readFile();
		}
		return bytes;
	}

	public int getChunkSize() {
		return chunkSize;
	}

	public void setChunkSize(int bufferSize) {
		this.chunkSize = bufferSize;
		this.readSize=BUFFER_SIZE-BUFFER_SIZE%bufferSize; 
	}
	public void setBufferChunks(int chunkSize, int numberOfChunks) {
		this.chunkSize=chunkSize;
		this.readSize=chunkSize * numberOfChunks;
	}
	
	public boolean isEOS() {
		return streamPosition + getAbsoluteFilePosition() >= streamSize || job.getStatus()==Job.STATUS_ABORTED;
	}
	public long getStreamSize() {
		return streamSize;
	}
	private void setJobSize() {
		job.setSize((int)(streamSize/Job.KB));
	}
	public void setStreamSize(long streamSize) {
		this.streamSize = streamSize;
		setJobSize();
	}
	public void  close() {
		if (closeFile()) {
			reset();
		}
	}
	private boolean closeFile() {
		try {
			if (fileChannel != null) {
				fileChannel.close();
				//System.out.println("Closed: " + files[currentIndex]);
				//fileIn.close();
				return true;
			}
		} catch (IOException e) {
		}
		return false;
	}
	public File[] getFiles() {
		return files;
	}
	public Job getJob() {
		return job;
	}
	public void setJob(Job job) {
		this.job=job;
		for (int i=0;i<files.length;i++) {
			streamSize+=files[i].length();
			if (i==0) {
				job.setDescription(files[0].getAbsolutePath());
			}
		}
		setJobSize();
		job.setPosition(0);
	}
	public String toString() {
		StringBuffer sb=new StringBuffer();
		String divider="";
		for (int i=0;i<files.length;i++) {
			sb.append(divider);
			sb.append(files[i].getAbsolutePath());
			divider=", ";
			
		}
		return sb.toString();
	}
	public long getFilePos() {
		return filePosition - fileBuffer.capacity() + fileBuffer.position();
	}
	
}
